/* [KSC4]-------------------------------------------------------------------[]
 *
 *  KTIMER.C - Countdown Timer for Kiko's Simple Timer/Counter, version 1.2
 *             By Marco "Kiko" Carnut <kiko at postcogito dot org>, May-2007
 *
 *  Description: The program multiplexes a 4-digit array of 8-digit segments,
 *  ------------ reads and debounces a 3-key keypad, implementing three
 *  functions:
 *
 *  1. an adjustable countdown timer that beeps for 30 seconds after the timer
 *     period expires. It also ticks the last 10 seconds.
 *  2. an upcounting chronometer with one-second granularity
 *  3. a hand counter, upcounting on keypress
 *
 *  This version is geared towards single-digit 8-segment displays with the
 *  standard pinout (10 pin, common anode is the middle pin in each row).
 *  It emulates the colon by having the last two digits upside down.
 *
 *  The software uses power down mode to implement an "auto-shutdown" feature,
 *  so the device has no on/off button. The keyboard circuit along with the
 *  ports configuration produce a pulse that triggers the AVR's external
 *  interrupt, waking it up.
 *
 *  It can be compiled with instrumentation counters stored in EEPROM that
 *  count how many times the device has been reset, how many times it woke
 *  up, the total amount of minutes counted, etc.
 *
 *  It has also support for receiving a limited number of commands over the
 *  serial port, including one to dump the instrumentation counters. The USART
 *  is set at 19,200 bps, 8 bits, no parity, 1 stop bit.
 *
 *  There are several compilation option defines in the configuration section
 *  that change several aspects of the code and allows selectively disabling
 *  uneeded features.
 *
 *  Portability: Tested on an ATtiny2313. Compiled under avr-gcc 3.4.4 and
 *  ------------ avr-libc 1.4.0.
 *
 *  License: This program is made available under the Creative Commons
 *  -------- Attribution-NonCommercial-ShareAlike version 2.5 license.
 *  Please see http://creativecommons.org/licenses/by-nc-sa/2.5/ for details.
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * []-----------------------------------------------------------------------[]
 */
 
#ifndef __AVR_ATtiny2313__
#error Hey, this code has been tested only on ATTiny2313
#endif // __AVR_ATtiny2313__

#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <avr/eeprom.h>

#define F_CPU                8000000L

#include <util/delay.h>

/* ------------------------ Configuration Defines ------------------------- */

// If QUELL_BUILTIN_CONFIG is defined, those bulit-in defines will not be
// defined. This is used by an automated build script to compute code size
// and resource occupation for different configurations.

#ifndef QUELL_BUILTIN_CONFIG

//#define HAVE_LN543GA

/* If this is defined, code will be generated to have the decimal points of
   the inner digits lit. */

#define HAVE_COLON

/* If this is defined, generates code to blink the colon when couting.
   Otherwise the colon will be fixed. The colon is supressed in the 'hand
   timer mode' */

#define DO_BLINK_COLON

/* Moves most variables to registers. Drastically reduces code size. */

#define USE_REGISTER_VARS

/* Adds support for the instrumentation counters */

#define HAVE_INSTRUMENTATION

/* Adds support for serial port */

#define HAVE_UART

/* If we should add a command to send us to the bootloader */

#define HAVE_ESCAPE_TO_BOOTLOADER

/* Adds support for a feature that halves the display refresh rate as the
   users keeps all three keys pressed. This is useful to demonstrate the
   multiplexing principle */

#define HAVE_MUX_DEMO

/* Save some extra code size by moving constants to EEPROM */

#define HAVE_CONSTANTS_ON_EEPROM

/* Should we have second-resultion adjustment of the timer initial value
   or only minute-resolution? */
   
#define HAVE_HIRES_ADJUST

/* If the HAVE_CALLIBRATION directive below is defined, code will be
   added to provide a callibration procedure for the main counter divisor,
   resulting in improved timing accuracy.

   The callibration procedure usually takes about five minutes and if
   correctly done, can achieve ~0.5% accuracy. It works like this:
   
   1. Get yourself another (hopefully more stable and accurate) clock to use
      as a reference
   2. Set the timer to 74:16
   3. Press and hold the ENTER button. After a few seconds holding it,
      the device will beep once a second to confirm the callibration mode
      has been activated. Do not release the ENTER key yet.
   4. Start your reference clock at the same time you release the ENTER key.
      Or, release the enter key when your already running reference clock
      goes past a reference point.
   5. When your reference clock reaches four minutes, press and hold the
      ENTER key.
   6. Release the enter key only when your reference clock says exactly 04:16
      (four minutes and 16 seconds) since the count started.
      
Your device is now calibrated. It beeps to confirm successful calibration.

If you need to abort the calibration process, you can either leave the
timer to expire (will take 70 minutes!) or remove the power to the system.
The callibration is held in RAM only, so it will be reverted back to the
default callibration when the device is reset.

Please notice that 0.5% accurracy still means about 30 seconds of error in the
full 99:59 minutes range. If your application needs more accuracy than that,
hook up a crystal.

*/

#define HAVE_CALLIBRATION

#endif // QUELL_BUILTIN_CONFIG

/* We can't have escape to bootloader if we don't have UART */

#ifndef HAVE_UART
#undef  HAVE_ESCAPE_TO_BOOTLOADER
#endif // HAVE_UART

/* Instrumentation counter locations */

#define ICNT_REBOOTS         0
#define ICNT_WAKEUPS         1
#define ICNT_MINUTES         2
#define ICNT_DUMPS           3

/* Serial transmission FIFO size. This MUST be a power of 2. */

#define MAXFIFOSIZE          16

/* -------------------- Hardware Configuration Defines -------------------- */

/* Ports leading to the driver transistors connecting to each display's
   common anode */

#define PORTD_MASK           (_BV(PD3) | _BV(PD4))
#define PORTA_MASK           (_BV(PA0) | _BV(PA1))

/* External device pins */

#define XTRNL_DEV_BIT        PD6
#define XTRNL_DEV_BV         _BV(XTRNL_DEV_BIT)

/* Buzzer */
#define BUZZER_DDR           DDRD
#define BUZZER_PORT          PORTD
#define BUZZER_BIT           PD5
#define BUZZER_BV            _BV(BUZZER_BIT)

/* Keyboard PINs and other definitions */

#ifdef HAVE_LN543GA

#define KBD_DEC              _BV(PB5)
#define KBD_INC              _BV(PB6)
#define KBD_ENTER            _BV(PB7)

#else

#define KBD_DEC              _BV(PB1)
#define KBD_INC              _BV(PB2)
#define KBD_ENTER            _BV(PB3)

#endif // HAVE_LN543GA

#define KBD_NKEYS            8
#define KBD_PASSES           6

#define KBD_MASK             (KBD_INC|KBD_DEC|KBD_ENTER)

/* --------------------------- Utility Defines ---------------------------- */

#ifdef USE_REGISTER_VARS
#define _r(a) asm(a)
#define DREG                 12
#define DREG_STR             "r12"
#define DREG_PTR             ((unsigned char *)(DREG))
#else
#define register
#define _r(a)
#define DREG_PTR             ((unsigned char *)(&display))
#endif // USE_REGISTER_VARS

#define TRUE                 1
#define FALSE                0

#ifdef HAVE_CONSTANTS_ON_EEPROM
#define STORAGE              EEMEM
#define storage_read_byte(a) eeprom_read_byte(a)
#define storage_read_word(a) eeprom_read_word((uint16_t *)(a))
#else
#define STORAGE              PROGMEM
#define storage_read_byte(a) pgm_read_byte(a)
#define storage_read_word(a) pgm_read_word(a)
#endif // HAVE_CONSTANTS_ON_EEPROM

/* --------------------- Global Variables and Declarations ---------------- */

/* This macro maps segments to specific bit assigments. It is used below
   to handle different display models/hookups */

#define LONGBIN2UCHAR(x,a,b,c,d,e,f,g,o)        \
    (unsigned char)(                            \
    ~ (                                         \
        ( (x) & 0x10000000 ? (1<<(a)) : 0 ) |   \
        ( (x) & 0x01000000 ? (1<<(b)) : 0 ) |   \
        ( (x) & 0x00100000 ? (1<<(c)) : 0 ) |   \
        ( (x) & 0x00010000 ? (1<<(d)) : 0 ) |   \
        ( (x) & 0x00001000 ? (1<<(e)) : 0 ) |   \
        ( (x) & 0x00000100 ? (1<<(f)) : 0 ) |   \
        ( (x) & 0x00000010 ? (1<<(g)) : 0 ) |   \
        ( (x) & 0x00000001 ? (1<<(o)) : 0 )     \
      )                                         \
    )

/* Those macros describe the hookup of the display segments, telling which
   PORTB bit corresponds to which segment when using different display models
   or hookup scenarios. The "SG"s are the "segment (display) groups". For now
   there are only two defined: SG1 for the upright mounted displays and SG2
   for the upside down mounted displays.*/

#ifdef HAVE_LN543GA

#define EXTRAVERSION "-LN543"

// This is for a different hookup using a LN543GA ":0.0:0.0:0.0:" (4-digit
// plus decimal points & colons) display module I found in a surplus store.

#define SG1(x) LONGBIN2UCHAR((x),0,4,5,7,1,3,6,2)
#define SG2(x) LONGBIN2UCHAR((x),7,3,2,0,6,4,1,5)

#else 

// This is for the common 7-segment displays

#define SG1(x) LONGBIN2UCHAR((x),5,0,2,4,3,6,7,1)
#define SG2(x) LONGBIN2UCHAR((x),5,0,2,4,3,6,1,7)

#endif // HAVE_LN543GA

// This defines the "font" glyphs, i.e.,        |   aaa
// how named characters are translated to       |  f   b
// LED display segments. Here, 1 means          |  f   b
// LIT, 0 means OFF. The diagram to the         |   ggg
// right shows the classic naming segment       |  e   c
// naming convention, with the slight           |  e   c
// difference that "o" means the decimal point. |   ddd  o

//                      abcdefgo
#define FONT_DIGIT0   0x11111100L
#define FONT_DIGIT1   0x01100000L
#define FONT_DIGIT2   0x11011010L
#define FONT_DIGIT3   0x11110010L
#define FONT_DIGIT4   0x01100110L
#define FONT_DIGIT5   0x10110110L
#define FONT_DIGIT6   0x10111110L
#define FONT_DIGIT7   0x11100000L
#define FONT_DIGIT8   0x11111110L
#define FONT_DIGIT9   0x11110110L
#define FONT_SPACE    0x00000000L
#define FONT_DP       0x00000001L

/* Maps integer values to specific "characters" or glyphs for the upright
   oriented display group, so we use the "SG1" macros. */

unsigned char STORAGE int2segm0[] = {
    SG1(FONT_DIGIT0),SG1(FONT_DIGIT1),SG1(FONT_DIGIT2),SG1(FONT_DIGIT3),
    SG1(FONT_DIGIT4),SG1(FONT_DIGIT5),SG1(FONT_DIGIT6),SG1(FONT_DIGIT7),
    SG1(FONT_DIGIT8),SG1(FONT_DIGIT9),SG1(FONT_SPACE)
};

/* Maps integer values to specific "characters" or glyphs for the upside-down
   oriented display group, so we use the "SG2" macros. */

unsigned char STORAGE int2segm1[] = {
    SG2(FONT_DIGIT0),SG2(FONT_DIGIT1),SG2(FONT_DIGIT2),SG2(FONT_DIGIT3),
    SG2(FONT_DIGIT4),SG2(FONT_DIGIT5),SG2(FONT_DIGIT6),SG2(FONT_DIGIT7),
    SG2(FONT_DIGIT8),SG2(FONT_DIGIT9),SG2(FONT_SPACE)
};

#define DOT1_MASK SG1(FONT_DP)
#define DOT2_MASK SG2(FONT_DP)

/* General status flags. Those are actually stored in the GPIOR0 register
   because it can be accessed by the sbi/cbi instructions, making code size
   a lot smaller */

#define FLAG0_ALARM           7  // Alarm is ringing
#define FLAG0_BEEPING         6  // Beep waveform being generated
#define FLAG0_RUNNING         5  // Upcount/downcount is running
#define FLAG0_CALLIBRATION    4  // This is a callibration run
#define FLAG0_HAS_INTERRUPT   3  // Timer interrupt just happened
#define FLAG0_COUNTER         2  // We're in hand counter mode
#define FLAG0_UPCOUNTING      1  // We're upcounting instead of downcounting
#define FLAG0_NZDIGIT         0  // Non-zero digit seen

#define FLAG1_SIMPLE_MODE     0  // Simple mode is enabled

// The display memory

register union {
    unsigned char  c[4];
    unsigned short s[2];
    unsigned long  l;
} display _r(DREG_STR);

register unsigned char digit         _r("r2"); // Current digit being muxed
register unsigned char ticks         _r("r3"); // Used to generate 1Hz
register unsigned char mux_mask      _r("r4"); // Slows down the mux

register unsigned char beeps         _r("r5"); // Number of beeps
register unsigned char cycle         _r("r6"); // beep cycle
register unsigned char cycle_length  _r("r7"); // how long to keep beeping

unsigned char kbd_count[KBD_NKEYS];             // Keypress counters
register unsigned char kbd_passes    _r("r8"); // scan passes done so far
register unsigned char count_on_key  _r("r9"); // Task to do on key release

//unsigned char repeat_count;                     // How long key pressed
#define repeat_count GPIOR2

register unsigned char sleep_level   _r("r10"); // How deep to sleep
register unsigned char idle_seconds  _r("r11"); // For how long been idle

#ifdef HAVE_UART

volatile unsigned char txfifo[MAXFIFOSIZE];

//volatile unsigned char txhead,txtail;

// A rather ugly trick to save some code space: we keep those variables
// in unused I/O registers. If you ever need those registers for something
// else, remember to undo this. :)

#define txhead USIDR
#define txtail OCR0B

#ifdef HAVE_INSTRUMENTATION
unsigned char xmit_cnt;
#endif // HAVE_INSTRUMENTATION

#endif // HAVE_UART

#ifdef HAVE_INSTRUMENTATION

//unsigned short minutes;

// Once again, the rather ugly trick of moving some variables to unused I/O
// registers to save code space. If you ever need to use ICR1 for something
// else, remember to undo this. :)

#define minutes OCR1B

unsigned char seconds;

unsigned short EEMEM icounters[] = {
   0,0,0,0,0,0,0
};

#endif // HAVE_INSTRUMENTATION

#ifdef HAVE_CALLIBRATION

//unsigned short cal_count;

// Yet again, the rather ugly trick of moving some variables to unused I/O
// registers to save code space. If you ever need to use ICR1 for something
// else, remember to undo this. :)

#define cal_count ICR1

#endif // HAVE_CALLIBRATION

#ifndef EXTRAVERSION
#define EXTRAVERSION "-GENERIC"
#endif

unsigned char version_string[] EEMEM = "v1.2" EXTRAVERSION "\r\n";

/* --------------------------- Function Prototypes ------------------------ */

void zero_display(void);
void clear_display(void);
void fill_display(unsigned char c);
void kill_alarm(void);
void handle_interrupt(void);
void inc_icounter(unsigned char c);
void count(unsigned char c, unsigned char i);
void start_running(void);
void stop_running(void);
void adjust(unsigned char direction);
void xmit(unsigned char c);
void xmit_nibble(unsigned char x);
void xmit_hex2(unsigned char c);
void xmit_hex4(unsigned short s);

/* ------------------------- Macro-Defined Functions ---------------------- */

#ifndef HAVE_HIRES_ADJUST
#define adjust(a) count(2,a)
#endif // HAVE_HIRES_ADJUST

/* ------------------------------ Main Program ---------------------------- */

int main(void)
{
    // Disable analog comparator to save power
    ACSR |= _BV(ACD);
    
    /* Port B pins are connected to the cathodes of the display segments
       (after a current limiting resistor) and to the keyboard switches */
    DDRB   = 0xFF;    
    PORTB  = 0xFF;
    
    /* Port D has two of the display mux transistors, INT0 (PD2) for waking
       up the part from power-down mode, the external device and the buzzer */
    DDRD   = PORTD_MASK | _BV(PD2) | XTRNL_DEV_BV;
    PORTD  = PORTD_MASK;

    /* Port A has the other two display mux transistors */
    DDRA   = PORTA_MASK;    
    PORTA  = PORTA_MASK;
        
    /* Timer0 Setup, used for the beep */
    
    TCCR0A = _BV(COM0B0) | _BV(WGM01);  // CTC mode, toggle on OCR0B match
    TCCR0B = 3;                         // CLK/64
    OCR0A  = 71;                        // 1,200 Hz beep frequency
    
    /* Timer1 Setup, provides time base for the display mux and 1Hz for
       the timer */
    TCCR1A = 0;
    TCCR1B = _BV(WGM12)             // CTC mode
           | 3;                     // CLK/64 = 125KHz
    OCR1A  = 499;                   // Output compare match each 4ms (250Hz)
                                    // NOTE: if you're using the onboard RC
                                    // oscillator, you may have to adjust this
                                    // to get good precision
    TIMSK  = _BV(OCIE1A);

#ifdef USE_REGISTER_VARS

    // Make sure things get initialized correctly

    zero_display();
    mux_mask=0;
    
#endif // USE_REGISTER_VARS
    
#ifdef HAVE_UART    

    /* UART Setup */
    UBRRL  = 25;                     // 19,200 bps with 0.2% error
    UCSRB |= _BV(RXEN) | _BV(TXEN); // Enable Transmitter and Receiver
    
#endif // HAVE_UART
        
#ifdef HAVE_INSTRUMENTATION

    inc_icounter(ICNT_REBOOTS);
    minutes=eeprom_read_word((uint16_t *)(&icounters[ICNT_MINUTES]));

#endif // HAVE_INSTRUMENTATION
    
    GPIOR1 = _BV(FLAG1_SIMPLE_MODE);
    
    for (;;) {
        sei();
        if (sleep_level==0xFF) {
            // Arm power-down sleep mode
            sleep_level=SLEEP_MODE_PWR_DOWN;         
            cli();                 // Disable interrupts
            DDRB   = KBD_MASK;     // Pull keyboard pins to ground so
            PORTB  = 0;            // keypresses will pull INT0 to ground
            DDRD   = 0;            // Configure all pins on port D as input
            PORTD  = _BV(PD2);     // Enable pullup on INT0 (PD2)
            DDRA   = 0;
            PORTA  = 0;
            GIMSK  = _BV(INT0);    // Enable Interrupt 0
            TIMSK  = 0;            // Disable Timer1 Interrupt
#ifdef HAVE_INSTRUMENTATION
            // Save Instrumentation minutes counter
            uint16_t *p=(uint16_t *)(&icounters[ICNT_MINUTES]);
            eeprom_write_word(p,minutes);
#endif // HAVE_INSTRUMENTATION
            sei();                 // Reenable interrupts
        }
        set_sleep_mode(sleep_level);  // Configure sleep mode
        sleep_mode();                 // Enter sleep mode
        asm("nop");
        // If we've been in power down mode, restore things
        if (sleep_level==SLEEP_MODE_PWR_DOWN) {
#ifdef HAVE_MUX_DEMO
            mux_mask=0;               // Full display mux speed
#endif // HAVE_MUX_DEMO
            DDRB   = 0xFF;            // Display pins are output again
            PORTB  = 0xFF;
            GIMSK  = 0;               // Disable INT0
            TIMSK  = _BV(OCIE1A);     // Reenable Timer1 interrupt
            // Reenable Port D outputs (display mux transistors,
            // keyboard pulldown and external device control)
            DDRD   = PORTD_MASK | _BV(PD2) | XTRNL_DEV_BV;
            PORTD  = PORTD_MASK;
            DDRA   = PORTA_MASK;
            PORTA  = PORTA_MASK;
            // Reset idle counter
            idle_seconds=0;           
            // Arm next sleep to be in idle mode
            sleep_level=SLEEP_MODE_IDLE;
#ifdef HAVE_INSTRUMENTATION
            // Register another wakeup in the instrumentation counters
            inc_icounter(ICNT_WAKEUPS);
#endif // HAVE_INSTRUMENTATION
            // In simple mode, we always reset the counter on wakeup
            // except if we're in counter mode
            if (GPIOR1 & _BV(FLAG1_SIMPLE_MODE)) {
                if (!(GPIOR0 & _BV(FLAG0_COUNTER))) {
                    zero_display();
                }
                // We also go back to coutdown mode
                GPIOR0 &= ~_BV(FLAG0_UPCOUNTING);
            }
        }
        if (GPIOR0 & _BV(FLAG0_HAS_INTERRUPT)) {
            GPIOR0 &= ~_BV(FLAG0_HAS_INTERRUPT);
            handle_interrupt();
        }
#ifdef HAVE_UART
        if (UCSRA & _BV(RXC)) {               // Have we receives something?
            unsigned char c=UDR;             // If so, get it
            idle_seconds=0;                   // reset the idle counter
#ifdef HAVE_INSTRUMENTATION
            if (c=='D'-'@') {                 // If it's CTRL-D, let's
                if (!xmit_cnt) {              // prepare the instrumentation
                    xmit_cnt=1;               // counter dump
                    inc_icounter(ICNT_DUMPS);
                }
            } else
#define HAVE_VERSION_STRING
#ifdef HAVE_VERSION_STRING
            if (c=='V'-'@') {
                EEAR = (unsigned char)version_string;
                for (;;) {
                    EECR |= _BV(EERE);
                    unsigned char c=EEDR;
                    if (!c) break;
                    xmit(c);
                    EEAR++;
                }
            } else
#endif // HAVE_VERSION_STRING
#endif // HAVE_INSTRUMENTATION
#ifdef HAVE_ESCAPE_TO_BOOTLOADER
            if (c==3) {                      // CTRL-C jumps to bootloader
                PORTD  = PORTD_MASK;         // Clear display
                PORTA  = PORTA_MASK;
                TIMSK  = 0;                  // Disable Timer1 Interrupt
                cli();                       // Globally disable interrupts
                // Jump to bootloader
                __asm__ __volatile__("rjmp bootstart");
            } else
#endif // HAVE_ESCAPE_TO_BOOTLOADER
            {
                xmit(c);                     // Otherwise just echo
            }
        }
#ifdef HAVE_INSTRUMENTATION
        if (xmit_cnt) {
            if (txhead==txtail) {      // If the xmit queue is empty
                xmit_hex2(xmit_cnt-1); // transmit the current instrumentation
                xmit(':');             // counter number, followed by colon,
                xmit(' ');             // space, and its value
                uint16_t *p=(uint16_t *)(&icounters[xmit_cnt-1]);
                xmit_hex4(eeprom_read_word(p));
                xmit('\r');            // and finish with CRLF
                xmit('\n');
                xmit_cnt++;
                xmit_cnt&=7;
            }
        }
#endif // HAVE_INSTRUMENTATION
#endif // HAVE_UART
    }
    return 0;
}

/* This is the main routine that implements most of the functionality,
   including display multiplexing, keyboard scanning and command handling,
   the 'user visible' countdown timer, beeping and so on. This is called
   each 4ms (250 times per second) */

void handle_interrupt(void)
{
    if (GPIOR0 & _BV(FLAG0_BEEPING)) {
        if (++cycle==cycle_length) { 
            cycle=0; 
            GPIOR0 &= ~_BV(FLAG0_BEEPING);
            beeps--;
            if (GPIOR0 & _BV(FLAG0_ALARM)) {
                clear_display();
            }
            BUZZER_DDR &= ~BUZZER_BV;    // Disable buzzer
        } else {
            BUZZER_DDR |=  BUZZER_BV;    // Reenable buzzer
        }
    }
    // Handles the user visible countdown timer functionality. We
    // enter this 'if' once a second
    if (++ticks==250) { 
        ticks=0;           
#ifdef HAVE_CALLIBRATION
        if (display.l==0x7040106 && 
            repeat_count>90) {
            GPIOR0 |= _BV(FLAG0_CALLIBRATION);
            GPIOR0 |= _BV(FLAG0_BEEPING);
            cal_count=0;
            cycle_length=40;
        } else if (GPIOR0 & _BV(FLAG0_CALLIBRATION)) cal_count++;
#endif // HAVE_CALLIBRATION        
        if (GPIOR0 & _BV(FLAG0_ALARM)) {
            if (beeps) {
                GPIOR0 |= _BV(FLAG0_BEEPING);
                zero_display();
            } else {
                kill_alarm();
                zero_display();
                sleep_level=0xFF;
            }
        } else if (GPIOR0 & _BV(FLAG0_RUNNING)) { // If the timer is running,
            idle_seconds=0;                         // we're not idle
            // count in the appropriate direction
            count(0,(GPIOR0 & _BV(FLAG0_UPCOUNTING)) ? 1 : -1); 
#ifdef HAVE_INSTRUMENTATION
            if (++seconds==60) {
                minutes++; seconds=0;
            }
#endif // HAVE_INSTRUMENTATION
            // Are we downcounting and near the end of the countdown?
            if (!(GPIOR0 & _BV(FLAG0_UPCOUNTING)) && display.s[1]==0) {
                if (display.s[0]<=256) {     // If we are in less than 10
                    GPIOR0 |= _BV(FLAG0_BEEPING);
                    if (display.s[0]==0) {   // see if the timer went to zero
                        GPIOR0 |= _BV(FLAG0_ALARM);
                        beeps=30; stop_running();
                        cycle_length=125;
                    } else {
                        // Otherwise arrange a 'tick' sound
                        cycle_length=2; beeps=1;
                    }
                }
            } 
        } else {
            if (++idle_seconds>19) {           // If we've been idle for too
               sleep_level=0xFF;                // long, arm power down
            }
        }        
    } 

    // This handles the display multiplexing. The colon generation and blinking
    // logic and leading-zero elimitation for counter mode are all here
    if (!(ticks&mux_mask)) {
        if (digit<4) {
            unsigned char mask=0xFF;
            unsigned char d=*(DREG_PTR+3-digit);
            if (d) GPIOR0 |= _BV(FLAG0_NZDIGIT);
            if (digit<2) {
                if (digit==0) {
                    PORTA &= ~_BV(PA0);
                } else {
#ifdef HAVE_COLON
                    if (d<10) mask=DOT1_MASK;
#endif // HAVE_COLON
                    PORTA = (PORTA|_BV(PA0)) & ~_BV(PA1);
                }
                d = storage_read_byte(int2segm0+d);
            } else {
                if (digit==2) {
#ifdef HAVE_COLON
                    if (d<10) mask=DOT2_MASK;
#endif // HAVE_COLON
                    PORTA |=  _BV(PA1);
                    PORTD &= ~_BV(PD3);
                } else {
                    PORTD = (PORTD|_BV(PD3)) & ~_BV(PD4);
                } 
                d = storage_read_byte(int2segm1+d);
            }
#ifdef HAVE_COLON
#ifdef DO_BLINK_COLON
            if ((GPIOR0 & _BV(FLAG0_RUNNING)) && ticks<125) mask=0xFF;
#endif // DO_BLINK_COLON
#endif // HAVE_COLON
            if (GPIOR0 & _BV(FLAG0_COUNTER)) {
                mask=0xFF;
                if (!(GPIOR0 & _BV(FLAG0_NZDIGIT)) && digit<3) d=0xFF;
            }
            PORTB = d & mask;
            digit++;
        } else {
            //
            // -- This handles keyboard scanning and commmand processing
            // 
            DDRB   = 0;            // Port B pins hooked to the keyboard are
            PORTB  = 0xFF;         // now inputs with pullups enabled
            // Port D is reconfigured to have PD2 pulling hard to ground
            // and the display is cleared (driver transistor's bases are
            // set high)
            DDRD  |=  PORTD_MASK | _BV(PD2);
            PORTD  =  (PORTD|PORTD_MASK) & ~_BV(PD2);
            // The two remaining transistors are driven high to disable
            // the display
            PORTA  =  PORTA_MASK;
            _delay_us(15); // Wait for ports to settle
            // By seeing which ones went down, we know which were pressed
            unsigned char current_keys = PINB;
            unsigned char n;
            // Accumulate those results for several passes
            for (n=0;n<KBD_NKEYS;n++) {
                if (!(current_keys&1)) kbd_count[n]++;
                current_keys>>=1;
            }
            // If we have done enough passes, let's see the results
            if (++kbd_passes==KBD_PASSES) {
                kbd_passes=0;
                unsigned char out=0;
                for (n=0;n<KBD_NKEYS;n++) {
                    out>>=1;
                    if (kbd_count[n]>=KBD_PASSES/2) {
                        out|=128; 
                    }
                    kbd_count[n]=0;
                }
                // Here we finally have in 'out' which keys are pressed,
                // so let's handle the commands
                if (!out) {
                    if (count_on_key) {                    
                        if (GPIOR0 & _BV(FLAG0_COUNTER)) {
                            count(0,count_on_key);
                        } else {
                            if (GPIOR0 & _BV(FLAG0_ALARM)) {
                                kill_alarm();
                            } else {
                                if (display.l) {
                                    if (GPIOR0 & _BV(FLAG0_RUNNING)) {
#ifdef HAVE_CALLIBRATION
                                        if (GPIOR0 & _BV(FLAG0_CALLIBRATION)) {
                                            GPIOR0 &= ~_BV(FLAG0_CALLIBRATION);
                                            unsigned short remainder=0;
                                            unsigned short new_ocr1a=0;
                                            unsigned short cc=cal_count;
                                            // The loop below computes the new
                                            // OCR1A value. It is the same as
                                            //
                                            // new_ocr1a=(OCR1A*cal_count)/256
                                            //
                                            // It is slower than using more 
                                            // sophisticated multiplication
                                            // and division routines, but it's
                                            // much smaller.
                                            for (;cc;cc--) {
                                                remainder+=OCR1A;
                                                while (remainder>=256) {
                                                   remainder-=256; 
                                                   new_ocr1a++;
                                                }
                                            }
                                            OCR1A=new_ocr1a;
                                            GPIOR0 |= _BV(FLAG0_BEEPING);
                                            cycle_length=130;
                                        }
#endif // HAVE_CALLIBRATION
                                        stop_running();
                                    } else {
                                        start_running();
                                    }
                                } else {
                                    // Starting the count with the display
                                    // at 00:00 enters upcount/chronometer
                                    // mode
                                    GPIOR0 |= _BV(FLAG0_UPCOUNTING);
                                    start_running();
                                }
                            }
                        }
                        count_on_key=0;
                    }
                    repeat_count=0;
                } else {
                    idle_seconds=0;
                    if (out==KBD_INC) {
                        if (!(GPIOR0 & _BV(FLAG0_COUNTER))) {
                            if (GPIOR0 & _BV(FLAG0_ALARM)) {
                                kill_alarm();
                            } else {
                                if (GPIOR0 & _BV(FLAG0_RUNNING)) {
                                    stop_running();
                                } else {
                                    adjust(1);
                                }
                            }
                        }
                    } else if (out==KBD_DEC) {
                        if (GPIOR0 & _BV(FLAG0_COUNTER)) {
                            count_on_key=-1;
                        } else {
                            if (GPIOR0 & _BV(FLAG0_ALARM)) {
                                kill_alarm();
                            } else {
                                if (GPIOR0 & _BV(FLAG0_RUNNING)) {
                                    stop_running();
                                } else {
                                    adjust(-1);
                                }
                            }
                        }
                    } else if (out&KBD_ENTER) {
                        if (out==KBD_ENTER) {
                            count_on_key=1;
                        }
                        else if (out==(KBD_INC|KBD_DEC|KBD_ENTER)) {
#ifdef HAVE_MUX_DEMO
                               // Pressing all three keys simultaneously
                               // halves the refresh rate. (It is restored
                               // back to its original value on wake up)
                               mux_mask++;
                               mux_mask<<=1;
                               mux_mask--;
                               if (mux_mask>0x1F) mux_mask=0;
#endif // HAVE_MUX_DEMO
                        }
                        else {
                            // Pressing '-' and 'ENTER' simultaneously enters
                            // hand counter mode if we're in advanced mode;
                            // if we keep it pressed for several seconds,
                            // we'll enable advanced mode, getting beeps to
                            // confirm.
                            // Pressing '+' and 'ENTER' simultaneously goes
                            // back to timer/chrono mode; if pressed for
                            // several seconds, we'll disable advanced mode
                            if (out&KBD_DEC) {
                                kill_alarm();
                                if (!(GPIOR1 & _BV(FLAG1_SIMPLE_MODE)))
                                    GPIOR0 |=  _BV(FLAG0_COUNTER);
                                if (repeat_count>90) {
                                    GPIOR1 &= ~_BV(FLAG1_SIMPLE_MODE);
                                    GPIOR0 |=  _BV(FLAG0_BEEPING);
                                    cycle_length=60;
                                }
                            } else {
                                GPIOR0 &= ~_BV(FLAG0_COUNTER);
                                if (repeat_count>90) {
                                    GPIOR1 |= _BV(FLAG1_SIMPLE_MODE);
                                    GPIOR0 |= _BV(FLAG0_BEEPING);
                                    cycle_length=60;
                                }
                            }
                            stop_running();
                            zero_display();
                        }
                    } else {
                        // Pressing '+' and '-' simultaneously zeroes the
                        // count and brings us back to timer/countdown mode
                        stop_running();
                        zero_display();
                        kill_alarm();
                        GPIOR0 &= ~_BV(FLAG0_UPCOUNTING);
                    }
                    if (!(repeat_count&128)) repeat_count++;
                    //if (repeat_count<100) repeat_count++;
                }
            }
            // We'll start a new frame. Clear the digit counter and the
            // 'non-zero digit seen' flag.
            GPIOR0 &= ~_BV(FLAG0_NZDIGIT);
            digit = 0;
            DDRB  = 0xFF;  // Segment pins are outputs again
        }
    }
    return;
}

void fill_display(unsigned char c)
{
    display.c[0]=display.c[1]=display.c[2]=display.c[3]=c;
}

void zero_display(void)
{
    fill_display(0);
}

void clear_display(void)
{
    fill_display(10);
}

void kill_alarm(void)
{
    GPIOR0 &= ~_BV(FLAG0_ALARM);
    GPIOR0 &= ~_BV(FLAG0_BEEPING);
    cycle=0;
    zero_display();
    DDRD  &= ~_BV(PD5);
}

#ifdef HAVE_INSTRUMENTATION

void inc_icounter(unsigned char c)
{
    uint16_t *address=(uint16_t *)(&icounters[c]);
    unsigned short val=eeprom_read_word(address);
    eeprom_write_word(address,val+1);
}

#endif // HAVE_INSTRUMENTATION

void count(unsigned char c, unsigned char i)
{
    unsigned char *p=DREG_PTR+c;
    for (;;) {
        unsigned char limit=9;
        if (!(GPIOR0 & _BV(FLAG0_COUNTER)) && p==DREG_PTR+1) limit=5;
        if (((*p)+=i)<=limit) break;
        *p=(i&8)?limit:0;
        if (p++==DREG_PTR+3) break;
    }
}

#ifdef HAVE_HIRES_ADJUST

void adjust(unsigned char direction)
{
    if (GPIOR1 & _BV(FLAG1_SIMPLE_MODE)) {
        count(2,direction); return;
    }
    if (repeat_count<10) {
        count(0,direction);
    } else if (repeat_count<33) {
        count(1,direction);
    } else {
        count(2,direction);
    }
}

#endif // HAVE_HIRES_ADJUST

void start_running(void)
{
    GPIOR0 |= _BV(FLAG0_RUNNING);
    PORTD  |= XTRNL_DEV_BV;  
    ticks=0;
}

void stop_running(void)
{
    GPIOR0 &= ~_BV(FLAG0_RUNNING);
    GPIOR0 &= ~_BV(FLAG0_CALLIBRATION);
    PORTD  &= ~XTRNL_DEV_BV;
}

/* ---------------------------- Interrupt Handlers ------------------------ */

/* Empty interrupt handler for INT0 -- this is used just to wake up the
   MCU then a key is pressed in power down mode */

EMPTY_INTERRUPT(SIG_INT0);

/* Handles Timer1 Comparator Match. Enables a flag telling the mainloop
   that it is time to handle the main chores (display multiplexing,
   keyboard scan, and so on) */

void SIG_TIMER1_COMPA(void) __attribute__((naked));
void SIG_TIMER1_COMPA(void)
{
    GPIOR0 |= _BV(FLAG0_HAS_INTERRUPT);
    asm volatile ("reti");
}

/* ---------------------------- Serial Port Support ----------------------- */

#ifdef HAVE_UART

/* USART Data Register Empty Interrupt Handler. Fetches anoter byte from the
   transmit queue and sends to the USART Data Register (UDR) for transmission.
   If the queue is empty, it disables the UDR Empty Interrupt. */

SIGNAL(SIG_USART0_DATA)
{
    if (txhead==txtail) {
        UCSRB &= ~_BV(UDRIE);
    } else {
        UDR = txfifo[txhead++];
        txhead &= MAXFIFOSIZE-1;
    }
}

/* Enqueues a byte for later transmission and enables the UDR Data Register
   Interrupt, causing the almost immediate execution of the transmission
   routine above */

void xmit(unsigned char c)
{
    txfifo[txtail++]=c;
    txtail &= MAXFIFOSIZE-1;
    UCSRB |= _BV(UDRIE);
}

/* Converts the lower four bits in an integer to an hex digit in the range
   [0-9A-F] and transmits it */

void xmit_nibble(unsigned char x)
{
    x &= 15;
    if (x>9) x+=7;
    xmit(x+'0');
}

/* Transmits an 8-bit integer as two hex digits */

void xmit_hex2(unsigned char c)
{
    xmit_nibble(c>>4); xmit_nibble(c);
}

/* Transmits a 16-bit integer as four hex digits */

void xmit_hex4(unsigned short s)
{
    union {
        unsigned char c[2];
        unsigned short s;
    } u;
    u.s=s;
    xmit_hex2(u.c[1]); xmit_hex2(u.c[0]);
}

#endif // HAVE_UART
